import os
import csv
import json
import pickle
import shutil

import yaml
import isddc_exception as ie


class FileSystem:

    def __init__(self, root: str):
        self.root: str = gendir(root)
        self.paths: dict = fetch_deep_path(self.root)
        self.objs = {}

        for rel_path, abs_path in self.paths.items():
            if isfile(abs_path):
                self.objs[rel_path] = FileSystem.File(abs_path)
            elif isdir(abs_path):
                self.objs[rel_path] = FileSystem.Directory(abs_path)
            else:
                self.objs[rel_path] = abs_path

    def __getitem__(self, rel_path: str):
        try:
            return self.objs[rel_path]
        except KeyError:
            return FileSystem.FalsePath(rel_path, joinpath(self.root, rel_path), self)

    def __delitem__(self, rel_path: str):
        if rel_path not in self.objs:
            raise ie.PathNotFoundError(joinpath(self.root, rel_path))
        del self.paths[rel_path]
        self.objs.pop(rel_path).delete()

    def __iter__(self):
        yield from self.objs.items()

    class FalsePath:

        def __init__(self, rel_path: str, abs_path: str, fs: 'FileSystem'):
            self.relpath = rel_path
            self.fs = fs
            self.fs_root: str = fs.root
            self.fs_objs: dict = fs.objs
            self.fs_paths: dict = fs.paths

        def __getattribute__(self, name: str):
            if name not in FileSystem.class_mapping:
                return super().__getattribute__(name)
            if self.relpath in self.fs_objs:
                return getattr(self.fs_objs[self.relpath], name)
            raise ie.PathNotFoundError(self.path)

        def __setattr__(self, name: str, value):
            if name not in ['content', 'subdirectory']:
                super().__setattr__(name, value)
            elif self.relpath in self.fs_objs:
                setattr(self.fs_objs[self.relpath], name, value)
            else:
                obj = FileSystem.class_mapping[name](self.path)

                dir_name: str = dirname(self.path)
                exists(dir_name) or makedirs(dir_name)

                dirs: list = dir_name.replace(self.fs_root, '').split(os.sep)[1:]
                for index in range(1, len(dirs) + 1):
                    rel_path = ''
                    for i in range(index):
                        rel_path += f'{dirs[i]}/'
                    rel_path = rel_path[:-1]

                    abs_path: str = joinpath(self.fs_root, rel_path)
                    self.fs_objs[rel_path] = FileSystem.Directory(abs_path)
                    self.fs_paths[rel_path] = abs_path

                setattr(obj, name, value)
                self.fs_objs[self.relpath] = obj
                self.fs_paths[self.relpath] = self.path

        def __delattr__(self, name: str):
            if name not in FileSystem.class_mapping:
                super().__delattr__(name)
            elif self.relpath in self.fs_objs:
                delattr(self.fs_objs[self.relpath], name)
            else:
                raise ie.PathNotFoundError(self.path)

        def __str__(self):
            return f'FalsePath({self.path})'

        @property
        def name(self) -> str:
            return basename(self.path)

    class File:

        def __init__(self, path: str):
            self.path = path

        def __str__(self):
            return f'File({self.path})'

        def __iter__(self):
            content = self.read()
            if isinstance(content, str):
                yield from content.split('\n')
            elif isinstance(content, list):
                yield from content
            elif isinstance(content, dict):
                yield from content.items()
            else:
                raise TypeError(f'Type {type(content).__name__} do not iteration.')

        def read(self, ftype: str = None):
            return Filetor(self.path, ftype=ftype)

        def write(self, content, *, ftype: str = None):
            Filetor(self.path, content, ftype=ftype)

        def clear(self):
            f = open(self.path, 'w', encoding='UTF-8')
            f.write('')
            f.close()

        content = property(read, write, clear)

        @property
        def name(self) -> str:
            return basename(self.path)

        def append(self, data):
            raise NotImplementedError

        @property
        def handle(self):
            raise NotImplementedError

        @property
        def size(self) -> int:
            return getsize(self.path)

        @property
        def atime(self):
            return getatime(self.path)

        @property
        def mtime(self) -> float:
            return getmtime(self.path)

        @property
        def ctime(self):
            return getctime(self.path)

        def delete(self):
            remove(self.path)

    class Directory:

        def __init__(self, path: str):
            self.path = path

        def __str__(self):
            return f'Directory({self.path})'

        def __iter__(self):
            yield from self.get_subdirectory()

        @property
        def name(self) -> str:
            return basename(self.path)

        def get_subdirectory(self) -> list:
            return listdir(self.path)

        def set_subdirectory(self, subdirectory: list):
            raise NotImplementedError

        def del_subdirectory(self):
            raise NotImplementedError

        def delete(self):
            raise NotImplementedError

        subdirectory = property(
            get_subdirectory, set_subdirectory, del_subdirectory)

    class_mapping = {}
    extract_method = lambda x: x[0] != '_' and x not in ['name', 'path']

    for method in filter(extract_method, dir(File(None))):
        class_mapping[method] = File
    for method in filter(extract_method, dir(Directory(None))):
        class_mapping[method] = Directory


class FileDataOperator:

    def __init__(self, root: str):
        self.root: str = gendir(root)
        self.path: dict = fetch_deep_path(self.root)

    def __getitem__(self, rel_path: str):
        abs_path: str = self.check_path_is_exists(rel_path)

        if not exists(abs_path):
            raise ie.PathIsDeletedError(abs_path)

        if isfile(abs_path):
            return Filetor(abs_path)

        return listdir(abs_path)

    def __setitem__(self, rel_path: str, data):
        if rel_path[0] == '/':
            rel_path = rel_path[1:]

        abs_path: str = joinpath(self.root, rel_path)
        Filetor(genbasedir(abs_path), data)

        self.path[rel_path] = abs_path

    def __delitem__(self, rel_path: str):
        abs_path: str = self.check_path_is_exists(rel_path)
        del self.path[rel_path]

        if isfile(abs_path):
            remove(abs_path)
        else:
            rmtree(abs_path, ignore_errors=True)

    def __getattr__(self, rel_path: str):
        return self[rel_path]

    def __setattr__(self, rel_path: str, data):
        if rel_path in ('root', 'path'):
            return super().__setattr__(rel_path, data)
        self[rel_path] = data

    def __delattr__(self, rel_path):
        del self[rel_path]

    def __getstate__(self):
        return self.root, self.path

    def __setstate__(self, state):
        self.root, self.path = state

    def __iter__(self):
        yield from self.path.items()

    def __str__(self):
        return f'{type(self).__name__}({self.root})'

    def check_path_is_exists(self, rel_path: str) -> str:
        if rel_path not in self.path:
            raise ie.PathNotFoundError(joinpath(self.root, rel_path))
        return self.path[rel_path]


class Filetor:
    def __new__(
            cls,
            file: str,
            data: ... = None,
            *,
            ftype: 'enum(json, yaml, obj, txt, csv)' = None,
            encoding: str = 'UTF-8',
            **kw
    ):
        mode = 'r' if data is None else 'w'

        if ftype is None:
            ftype: str = file.rsplit('.', 1)[-1]

        if ftype == 'obj':
            mode += 'b'
            encoding = None

        newline = '' if ftype == 'csv' else None

        with open(file, mode, encoding=encoding, newline=newline) as fp:
            try:
                return getattr(cls, f'_for_{ftype}_')(fp, data, **kw)
            except AttributeError:
                return cls._for_text_(fp, data, **kw)

    @staticmethod
    def _for_json_(fp: open, data=None, **kw):
        if data is None:
            return json.load(fp, **kw)
        return json.dump(data, fp, **kw)

    @staticmethod
    def _for_yaml_(fp: open, data=None, many: bool = False, **kw):
        if data:
            if many:
                return yaml.safe_dump_all(data, fp, **kw)
            return yaml.safe_dump(data, fp, **kw)
        if many:
            return list(yaml.safe_load_all(fp))
        return yaml.safe_load(fp)

    @staticmethod
    def _for_obj_(fp: open, obj=None, **kw):
        if obj is None:
            return pickle.load(fp, **kw)
        return pickle.dump(obj, fp, **kw)

    @staticmethod
    def _for_text_(fp: open, data=None, **kw):
        if data is None:
            return fp.read(**kw)
        return fp.write(str(data))

    @staticmethod
    def _for_csv_(fp: open, data=None, dialect: str = 'excel', **kw):
        if dialect not in csv.list_dialects():
            csv.register_dialect(dialect, delimiter=dialect, quoting=csv.QUOTE_ALL)

        if data is None:
            cf = csv.DictReader(fp, dialect=dialect)
            return list(cf)

        if not data:
            raise ie.CSVFileIsEmpptyOrUntitledError(
                'The csv content cannot be empty and must include headers.')

        cf = csv.DictWriter(fp, data[0].keys(), dialect=dialect)
        cf.writeheader()
        cf.writerows(data)

    _for_yml_ = _for_yaml_
    _for_txt_ = _for_text_


def fetch_deep_path(
        path: str = None,
        _paths: 'Not param' = None,
        _root_path: 'Not param' = None,
) -> dict:
    paths: dict = _paths or {}
    root_path: str = _root_path or abspath(path)

    if path == root_path:
        paths['_root'] = path

    for name in listdir(path):
        abs_path: str = joinpath2abs(path, name)
        rel_path: str = abs_path.replace(root_path, '')[1:].replace('\\', '/')
        paths[rel_path] = abs_path

        if isdir(abs_path):
            fetch_deep_path(abs_path, paths, root_path)

    return paths


def joinpath2abs(*a) -> str:
    return abspath(joinpath(*a))


def dirname(path: str, *, level: int = 1) -> str:
    path: str = abspath(path)
    for n in range(level):
        path = os.path.dirname(path)
    return path


def gendir(*a) -> str:
    dir_: str = joinpath2abs(*a)
    exists(dir_) or makedirs(dir_)
    return dir_


def genbasedir(*a) -> str:
    path: str = joinpath2abs(*a)
    dirpath: str = os.path.dirname(path)
    exists(dirpath) or makedirs(dirpath)
    return path


joinpath = os.path.join
abspath  = os.path.abspath
relpath  = os.path.relpath
basename = os.path.basename

isabs   = os.path.isabs
isdir   = os.path.isdir
isfile  = os.path.isfile
islink  = os.path.islink
ismount = os.path.ismount
exists  = os.path.exists

rename = os.rename
remove = os.remove
rmdir  = os.rmdir
rmtree = shutil.rmtree

mkdir    = os.mkdir
makedirs = os.makedirs
listdir  = os.listdir

getsize  = os.path.getsize
getatime = os.path.getatime
getmtime = os.path.getmtime
getctime = os.path.getctime
